@@ -50,12 +50,11 @@
 #include <unistd.h>
 #include <errno.h>
 #include <sys/ioctl.h>
-#include <scsi/scsi.h>
 #include <scsi/sg.h>
 #include <scsi/sg_lib.h>
 #include <scsi/sg_cmds.h>
 
-static char * version_str = "1.17 20050719";
+static char * version_str = "1.21 20051027";
 
 
 #define SENSE_BUFF_LEN 32       /* Arbitrary, could be larger */
@@ -87,7 +86,11 @@ static char * version_str = "1.17 20050719";
 #define REPORT_LUNS_CMDLEN 12
 #define MAINTENANCE_IN_CMD 0xa3
 #define MAINTENANCE_IN_CMDLEN 12
-#define REPORT_TGT_PRT_GRP_SA 0x0a
+#define REPORT_TGT_PRT_GRP_SA 0xa
+#define REPORT_DEVICE_IDENTIFIER_SA 0x5
+#define MAINTENANCE_OUT_CMD 0xa4
+#define MAINTENANCE_OUT_CMDLEN 12
+#define SET_DEVICE_IDENTIFIER_SA 0x6
 #define LOG_SENSE_CMD     0x4d
 #define LOG_SENSE_CMDLEN  10
 #define LOG_SELECT_CMD     0x4c
@@ -125,10 +128,11 @@ const char * sg_cmds_version()
 int sg_ll_inquiry(int sg_fd, int cmddt, int evpd, int pg_op, 
                   void * resp, int mx_resp_len, int noisy, int verbose)
 {
-    int res, k;
+    int res, k, got;
     unsigned char inqCmdBlk[INQUIRY_CMDLEN] = {INQUIRY_CMD, 0, 0, 0, 0, 0};
     unsigned char sense_b[SENSE_BUFF_LEN];
     struct sg_io_hdr io_hdr;
+    unsigned char * up;
 
     if (cmddt)
         inqCmdBlk[1] |= 2;
@@ -146,6 +150,12 @@ int sg_ll_inquiry(int sg_fd, int cmddt, int evpd, int pg_op,
             fprintf(sg_warnings_str, "%02x ", inqCmdBlk[k]);
         fprintf(sg_warnings_str, "\n");
     }
+    if (resp && (mx_resp_len > 0)) {
+        up = resp;
+        up[0] = 0x7f;   /* defensive prefill */
+        if (mx_resp_len > 4)
+            up[4] = 0;
+    }
     memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
     memset(sense_b, 0, sizeof(sense_b));
     io_hdr.interface_id = 'S';
@@ -170,11 +180,19 @@ int sg_ll_inquiry(int sg_fd, int cmddt, int evpd, int pg_op,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Inquiry", &io_hdr);
+            sg_chk_n_print3("Inquiry", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
+        got = io_hdr.dxfer_len - io_hdr.resid;
+        if (got < 4) {
+            if (verbose)
+                fprintf(sg_warnings_str, "inquiry: got too few (%d) "
+                        "bytes\n", got);
+            return -2;
+        }
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    inquiry: resid=%d\n", io_hdr.resid);
+            fprintf(sg_warnings_str, "    inquiry: resid=%d (got %d "
+                    "bytes)\n", io_hdr.resid, got);
         return 0;
     default:
         if (noisy || verbose) {
@@ -188,7 +206,7 @@ int sg_ll_inquiry(int sg_fd, int cmddt, int evpd, int pg_op,
                          pg_op);
             else
                 snprintf(ebuff, EBUFF_SZ, "Inquiry error, [standard]");
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -2;
     }
@@ -199,7 +217,7 @@ int sg_ll_inquiry(int sg_fd, int cmddt, int evpd, int pg_op,
 int sg_simple_inquiry(int sg_fd, struct sg_simple_inquiry_resp * inq_data,
                       int noisy, int verbose)
 {
-    int res, k;
+    int res, k, got;
     unsigned char inqCmdBlk[INQUIRY_CMDLEN] = {INQUIRY_CMD, 0, 0, 0, 0, 0};
     unsigned char sense_b[SENSE_BUFF_LEN];
     struct sg_io_hdr io_hdr;
@@ -219,6 +237,8 @@ int sg_simple_inquiry(int sg_fd, struct sg_simple_inquiry_resp * inq_data,
             fprintf(sg_warnings_str, "%02x ", inqCmdBlk[k]);
         fprintf(sg_warnings_str, "\n");
     }
+    memset(inq_resp, 0, sizeof(inq_resp));
+    inq_resp[0] = 0x7f; /* defensive prefill */
     memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
     memset(sense_b, 0, sizeof(sense_b));
     io_hdr.interface_id = 'S';
@@ -243,9 +263,16 @@ int sg_simple_inquiry(int sg_fd, struct sg_simple_inquiry_resp * inq_data,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Inquiry", &io_hdr);
+            sg_chk_n_print3("Inquiry", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
+        got = io_hdr.dxfer_len - io_hdr.resid;
+        if (got < 4) {
+            if (verbose)
+                fprintf(sg_warnings_str, "inquiry: got too few (%d) "
+                        "bytes\n", got);
+            return -2;
+        }
         if (inq_data) {
             inq_data->peripheral_qualifier = (inq_resp[0] >> 5) & 0x7;
             inq_data->peripheral_type = inq_resp[0] & 0x1f;
@@ -260,14 +287,15 @@ int sg_simple_inquiry(int sg_fd, struct sg_simple_inquiry_resp * inq_data,
             memcpy(inq_data->revision, inq_resp + 32, 4);
         }
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    inquiry: resid=%d\n", io_hdr.resid);
+            fprintf(sg_warnings_str, "    inquiry: resid=%d (got %d "
+                    "bytes)\n", io_hdr.resid, got);
         return 0;
     default:
-        if (noisy) {
+        if (noisy || verbose) {
             char ebuff[EBUFF_SZ];
 
             snprintf(ebuff, EBUFF_SZ, "Inquiry error ");
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -2;
     }
@@ -275,8 +303,11 @@ int sg_simple_inquiry(int sg_fd, struct sg_simple_inquiry_resp * inq_data,
 
 /* Invokes a SCSI TEST UNIT READY command.
  * 'pack_id' is just for diagnostics, safe to set to 0.
+ * Looks for progress indicator if 'progress' non-NULL;
+ * if found writes value [0..65535] else write -1.
  * Return of 0 -> success, -1 -> failure */
-int sg_ll_test_unit_ready(int sg_fd, int pack_id, int noisy, int verbose)
+int sg_ll_test_unit_ready_progress(int sg_fd, int pack_id, int * progress,
+                                   int noisy, int verbose)
 {
     int res, k;
     unsigned char turCmbBlk[TUR_CMDLEN] = {TUR_CMD, 0, 0, 0, 0, 0};
@@ -317,12 +348,26 @@ int sg_ll_test_unit_ready(int sg_fd, int pack_id, int noisy, int verbose)
     case SG_LIB_CAT_CLEAN:
         return 0;
     default:
+        if (progress) {
+            if (! sg_get_sense_progress_fld(sense_b, sizeof(sense_b),
+                                            progress))
+                *progress = -1;
+        }
         if (noisy || verbose)
-            sg_chk_n_print3("test unit ready", &io_hdr);
+            sg_chk_n_print3("test unit ready", &io_hdr, (verbose > 1));
         return -1;
     }
 }
 
+/* Invokes a SCSI TEST UNIT READY command.
+ * 'pack_id' is just for diagnostics, safe to set to 0.
+ * Return of 0 -> success, -1 -> failure */
+int sg_ll_test_unit_ready(int sg_fd, int pack_id, int noisy, int verbose)
+{
+    return sg_ll_test_unit_ready_progress(sg_fd, pack_id, NULL, noisy,
+                                          verbose);
+}
+
 /* Invokes a SCSI SYNCHRONIZE CACHE (10) command. Return of 0 -> success,
  * -1 -> failure, SG_LIB_CAT_MEDIA_CHANGED -> repeat, SG_LIB_CAT_INVALID_OP
  * -> cdb not supported, SG_LIB_CAT_IlLEGAL_REQ -> bad field in cdb */
@@ -388,11 +433,11 @@ int sg_ll_sync_cache_10(int sg_fd, int sync_nv, int immed, int group,
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("synchronize cache", &io_hdr);
+            sg_chk_n_print3("synchronize cache", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose)
-            sg_chk_n_print3("synchronize cache", &io_hdr);
+            sg_chk_n_print3("synchronize cache", &io_hdr, (verbose > 1));
         return -1;
     }
     return 0;
@@ -403,7 +448,7 @@ int sg_ll_sync_cache_10(int sg_fd, int sync_nv, int immed, int group,
  * -1 -> failure, SG_LIB_CAT_MEDIA_CHANGED -> repeat, SG_LIB_CAT_INVALID_OP
  *  -> cdb not supported, SG_LIB_CAT_IlLEGAL_REQ -> bad field in cdb */
 int sg_ll_readcap_16(int sg_fd, int pmi, unsigned long long llba, 
-                     void * resp, int mx_resp_len, int verbose)
+                     void * resp, int mx_resp_len, int noisy, int verbose)
 {
     int k, res;
     unsigned char rcCmdBlk[SERVICE_ACTION_IN_16_CMDLEN] = 
@@ -460,22 +505,24 @@ int sg_ll_readcap_16(int sg_fd, int pmi, unsigned long long llba,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (verbose)
-            sg_chk_n_print3("Read capacity (16)", &io_hdr);
+            sg_chk_n_print3("Read capacity (16)", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    read_capacity16: resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    read_capacity16: resid=%d (got"
+                    " %d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
     case SG_LIB_CAT_MEDIA_CHANGED:
         if (verbose > 1)
-            sg_chk_n_print3("READ CAPACITY 16 command error", &io_hdr);
+            sg_chk_n_print3("READ CAPACITY 16 command error", &io_hdr, 1);
         return res;
     default:
-        if (verbose)
-            sg_chk_n_print3("READ CAPACITY 16 command error", &io_hdr);
+        if (noisy || verbose)
+            sg_chk_n_print3("READ CAPACITY 16 command error", &io_hdr,
+                            (verbose > 1));
         return -1;
     }
 }
@@ -485,7 +532,7 @@ int sg_ll_readcap_16(int sg_fd, int pmi, unsigned long long llba,
  * -> media changed, SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb,
  * -1 -> other failure */
 int sg_ll_readcap_10(int sg_fd, int pmi, unsigned int lba, 
-                     void * resp, int mx_resp_len, int verbose)
+                     void * resp, int mx_resp_len, int noisy, int verbose)
 {
     int k, res;
     unsigned char rcCmdBlk[READ_CAPACITY_10_CMDLEN] =
@@ -532,22 +579,24 @@ int sg_ll_readcap_10(int sg_fd, int pmi, unsigned int lba,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (verbose)
-            sg_chk_n_print3("Read capacity (10)", &io_hdr);
+            sg_chk_n_print3("Read capacity (10)", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    read_capacity10: resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    read_capacity10: resid=%d (got"
+                    " %d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
     case SG_LIB_CAT_MEDIA_CHANGED:
         if (verbose > 1)
-            sg_chk_n_print3("READ CAPACITY 10 command error", &io_hdr);
+            sg_chk_n_print3("READ CAPACITY 10 command error", &io_hdr, 1);
         return res;
     default:
-        if (verbose)
-            sg_chk_n_print3("READ CAPACITY 10 command error", &io_hdr);
+        if (noisy || verbose)
+            sg_chk_n_print3("READ CAPACITY 10 command error", &io_hdr,
+                            (verbose > 1));
         return -1;
     }
 }
@@ -605,12 +654,13 @@ int sg_ll_mode_sense6(int sg_fd, int dbd, int pc, int pg_code, int sub_pg_code,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Mode sense (6)", &io_hdr);
+            sg_chk_n_print3("Mode sense (6)", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    mode sense (6): resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    mode sense (6): resid=%d (got "
+                    "%d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
         if (verbose > 2) {
             k = mx_resp_len - io_hdr.resid;
             if (k > 0) {
@@ -623,7 +673,7 @@ int sg_ll_mode_sense6(int sg_fd, int dbd, int pc, int pg_code, int sub_pg_code,
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("Mode sense (6) error", &io_hdr);
+            sg_chk_n_print3("Mode sense (6) error", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose) {
@@ -632,7 +682,7 @@ int sg_ll_mode_sense6(int sg_fd, int dbd, int pc, int pg_code, int sub_pg_code,
             snprintf(ebuff, EBUFF_SZ, "Mode sense (6) error, dbd=%d "
                     "pc=%d page_code=%x sub_page_code=%x\n     ", dbd,
                     pc, pg_code, sub_pg_code);
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -1;
     }
@@ -693,12 +743,13 @@ int sg_ll_mode_sense10(int sg_fd, int llbaa, int dbd, int pc, int pg_code,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Mode sense (10)", &io_hdr);
+            sg_chk_n_print3("Mode sense (10)", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    mode sense (10): resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    mode sense (10): resid=%d (got "
+                    "%d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
         if (verbose > 2) {
             k = mx_resp_len - io_hdr.resid;
             if (k > 0) {
@@ -711,7 +762,7 @@ int sg_ll_mode_sense10(int sg_fd, int llbaa, int dbd, int pc, int pg_code,
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("Mode sense (10) error", &io_hdr);
+            sg_chk_n_print3("Mode sense (10) error", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose) {
@@ -720,7 +771,7 @@ int sg_ll_mode_sense10(int sg_fd, int llbaa, int dbd, int pc, int pg_code,
             snprintf(ebuff, EBUFF_SZ, "Mode sense (10) error, dbd=%d "
                     "pc=%d page_code=%x sub_page_code=%x\n     ", dbd,
                     pc, pg_code, sub_pg_code);
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -1;
     }
@@ -781,14 +832,14 @@ int sg_ll_mode_select6(int sg_fd, int pf, int sp, void * paramp,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Mode select (6)", &io_hdr);
+            sg_chk_n_print3("Mode select (6)", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("Mode select (6) error", &io_hdr);
+            sg_chk_n_print3("Mode select (6) error", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose) {
@@ -796,7 +847,7 @@ int sg_ll_mode_select6(int sg_fd, int pf, int sp, void * paramp,
 
             snprintf(ebuff, EBUFF_SZ, "Mode select (6) error, pf=%d "
                     "sp=%d\n     ", pf, sp);
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -1;
     }
@@ -858,14 +909,14 @@ int sg_ll_mode_select10(int sg_fd, int pf, int sp, void * paramp,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Mode select (10)", &io_hdr);
+            sg_chk_n_print3("Mode select (10)", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("Mode select (10) error", &io_hdr);
+            sg_chk_n_print3("Mode select (10) error", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose) {
@@ -873,7 +924,7 @@ int sg_ll_mode_select10(int sg_fd, int pf, int sp, void * paramp,
 
             snprintf(ebuff, EBUFF_SZ, "Mode select (10) error, pf=%d "
                     "sp=%d\n     ", pf, sp);
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -1;
     }
@@ -1043,7 +1094,7 @@ int sg_get_mode_page_controls(int sg_fd, int mode6, int pg_code,
  * SG_LIB_CAT_INVALID_OP -> Request Sense not * supported??,
  * SG_LIB_CAT_ILEGAL_REQ -> bad field in cdb, -1 -> other failure */
 int sg_ll_request_sense(int sg_fd, int desc, void * resp, int mx_resp_len,
-                        int verbose)
+                        int noisy, int verbose)
 {
     int k, res;
     unsigned char rsCmdBlk[REQUEST_SENSE_CMDLEN] = 
@@ -1094,25 +1145,28 @@ int sg_ll_request_sense(int sg_fd, int desc, void * resp, int mx_resp_len,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (verbose)
-            sg_chk_n_print3("Request sense", &io_hdr);
+            sg_chk_n_print3("Request sense", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if ((mx_resp_len >= 8) && (io_hdr.resid > (mx_resp_len - 8))) {
-            fprintf(sg_warnings_str, "    request sense: resid=%d "
-                    "indicates response too short\n", io_hdr.resid);
+            if (verbose)
+                fprintf(sg_warnings_str, "    request sense: resid=%d "
+                        "indicates response too short\n", io_hdr.resid);
             return -1;
         } else if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    request sense: resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    request sense: resid=%d (got "
+                    "%d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("REQUEST SENSE command problem", &io_hdr);
+            sg_chk_n_print3("REQUEST SENSE command problem", &io_hdr, 1);
         return res;
     default:
-        if (verbose)
-            sg_chk_n_print3("REQUEST SENSE command problem", &io_hdr);
+        if (noisy || verbose)
+            sg_chk_n_print3("REQUEST SENSE command problem", &io_hdr,
+                            (verbose > 1));
         return -1;
     }
 }
@@ -1166,23 +1220,25 @@ int sg_ll_report_luns(int sg_fd, int select_report, void * resp,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Report luns", &io_hdr);
+            sg_chk_n_print3("Report luns", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    report_luns: resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    report luns: resid=%d (got "
+                    "%d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("REPORTS LUNS command error", &io_hdr);
+            sg_chk_n_print3("REPORTS LUNS command error", &io_hdr, 1);
         return res;
     case SG_LIB_CAT_MEDIA_CHANGED:
         return 2;
     default:
         if (noisy || verbose)
-            sg_chk_n_print3("REPORT LUNS command error", &io_hdr);
+            sg_chk_n_print3("REPORT LUNS command error", &io_hdr,
+                            (verbose > 1));
         return -1;
     }
 }
@@ -1243,17 +1299,18 @@ int sg_ll_log_sense(int sg_fd, int ppc, int sp, int pc, int pg_code,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Log sense", &io_hdr);
+            sg_chk_n_print3("Log sense", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    log_sense: resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    log sense: resid=%d (got "
+                    "%d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("log_sense error", &io_hdr);
+            sg_chk_n_print3("log_sense error", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose) {
@@ -1261,7 +1318,7 @@ int sg_ll_log_sense(int sg_fd, int ppc, int sp, int pc, int pg_code,
             snprintf(ebuff, EBUFF_SZ, "log_sense: ppc=%d, sp=%d, "
                      "pc=%d, page_code=%x, paramp=%x\n    ", ppc, sp, pc, 
                      pg_code, paramp);
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -1;
     }
@@ -1326,24 +1383,21 @@ int sg_ll_log_select(int sg_fd, int pcr, int sp, int pc,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Log select", &io_hdr);
+            sg_chk_n_print3("Log select", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
-        if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    log_select: resid=%d\n",
-                    io_hdr.resid);
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("log_select error", &io_hdr);
+            sg_chk_n_print3("log_select error", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose) {
             char ebuff[EBUFF_SZ];
             snprintf(ebuff, EBUFF_SZ, "log_select: pcr=%d, sp=%d, "
                      "pc=%d\n    ", pcr, sp, pc);
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -1;
     }
@@ -1398,21 +1452,24 @@ int sg_ll_report_tgt_prt_grp(int sg_fd, void * resp,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Report target port groups", &io_hdr);
+            sg_chk_n_print3("Report target port groups", &io_hdr,
+                            (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    report_tgt_prt_grp: resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    report_tgt_prt_grp: resid=%d (got "
+                    "%d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("REPORT TARGET PORT GROUPS", &io_hdr);
+            sg_chk_n_print3("REPORT TARGET PORT GROUPS", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose)
-            sg_chk_n_print3("REPORT TARGET PORT GROUPS command error", &io_hdr);
+            sg_chk_n_print3("REPORT TARGET PORT GROUPS command error",
+                            &io_hdr, (verbose > 1));
         return -1;
     }
 }
@@ -1474,21 +1531,22 @@ int sg_ll_send_diag(int sg_fd, int sf_code, int pf_bit, int sf_bit,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Send diagnostic, continuing", &io_hdr);
+            sg_chk_n_print3("Send diagnostic, continuing", &io_hdr,
+                            (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("SEND DIAGNOSTIC", &io_hdr);
+            sg_chk_n_print3("SEND DIAGNOSTIC", &io_hdr, 1);
         return res;
     default:
-        if (noisy) {
+        if (noisy || verbose) {
             char ebuff[EBUFF_SZ];
             snprintf(ebuff, EBUFF_SZ, "Send diagnostic error, sf_code=0x%x, "
                      "pf_bit=%d, sf_bit=%d ", sf_code, pf_bit, sf_bit);
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -1;
     }
@@ -1544,21 +1602,21 @@ int sg_ll_receive_diag(int sg_fd, int pcv, int pg_code, void * resp,
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
             sg_chk_n_print3("Receive diagnostics results, continuing",
-                            &io_hdr);
+                            &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("RECEIVE DIAGNOSTICS RESULTS", &io_hdr);
+            sg_chk_n_print3("RECEIVE DIAGNOSTICS RESULTS", &io_hdr, 1);
         return res;
     default:
         if (noisy) {
             char ebuff[EBUFF_SZ];
             snprintf(ebuff, EBUFF_SZ, "Receive diagnostics results error, "
                      "pcv=%d, page_code=%x ", pcv, pg_code);
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -1;
     }
@@ -1618,12 +1676,13 @@ int sg_ll_read_defect10(int sg_fd, int req_plist, int req_glist,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Read defect (10)", &io_hdr);
+            sg_chk_n_print3("Read defect (10)", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    read defect (10): resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    report defect(10): resid=%d (got "
+                    "%d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
         if (verbose > 2) {
             k = mx_resp_len - io_hdr.resid;
             if (k > 0) {
@@ -1636,7 +1695,7 @@ int sg_ll_read_defect10(int sg_fd, int req_plist, int req_glist,
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("Read defect (10) error", &io_hdr);
+            sg_chk_n_print3("Read defect (10) error", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose) {
@@ -1645,7 +1704,7 @@ int sg_ll_read_defect10(int sg_fd, int req_plist, int req_glist,
             snprintf(ebuff, EBUFF_SZ, "Read defect (10) error, req_plist=%d "
                     "req_glist=%d dl_format=%x\n     ", req_plist, req_glist,
                     dl_format);
-            sg_chk_n_print3(ebuff, &io_hdr);
+            sg_chk_n_print3(ebuff, &io_hdr, (verbose > 1));
         }
         return -1;
     }
@@ -1654,8 +1713,8 @@ int sg_ll_read_defect10(int sg_fd, int req_plist, int req_glist,
 /* Invokes a SCSI READ MEDIA SERIAL NUMBER command. Return of 0 -> success,
  * SG_LIB_CAT_INVALID_OP -> Read media serial number not supported,
  * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, -1 -> other failure */
-int sg_ll_read_media_serial_num(int sg_fd, void * resp,
-                                int mx_resp_len, int noisy, int verbose)
+int sg_ll_read_media_serial_num(int sg_fd, void * resp, int mx_resp_len,
+                                int noisy, int verbose)
 {
     int k, res;
     unsigned char rmsnCmdBlk[SERVICE_ACTION_IN_12_CMDLEN] =
@@ -1700,22 +1759,32 @@ int sg_ll_read_media_serial_num(int sg_fd, void * resp,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (noisy || verbose)
-            sg_chk_n_print3("Read media serial number", &io_hdr);
+            sg_chk_n_print3("Read media serial number", &io_hdr,
+                            (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    read_media_serial_num: resid=%d\n",
-                    io_hdr.resid);
+            fprintf(sg_warnings_str, "    read_media_serial_num: resid=%d ("
+                    "got %d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
+        if (verbose > 2) {
+            k = mx_resp_len - io_hdr.resid;
+            if (k > 0) {
+                fprintf(sg_warnings_str, "    report media serial number: "
+                        "response%s\n", (k > 256 ? ", first 256 bytes" : ""));
+                dStrHex(resp, (k > 256 ? 256 : k), -1);
+            }
+        }
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("READ MEDIA SERIAL NUMBER", &io_hdr);
+            sg_chk_n_print3("READ MEDIA SERIAL NUMBER", &io_hdr, 1);
         return res;
     default:
         if (noisy || verbose)
             sg_chk_n_print3("READ MEDIA SERIAL NUMBER command error",
-                            &io_hdr);
+                            &io_hdr, (verbose > 1));
         return -1;
     }
 }
@@ -1725,7 +1794,7 @@ int sg_ll_read_media_serial_num(int sg_fd, void * resp,
  * SG_LIB_CAT_INVALID_OP -> Start stop unit not supported,
  * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, -1 -> other failure */
 int sg_ll_start_stop_unit(int sg_fd, int immed, int power_cond,
-                          int loej, int start, int verbose)
+                          int loej, int start, int noisy, int verbose)
 {
     unsigned char ssuBlk[START_STOP_CMDLEN] = {START_STOP_CMD, 0, 0, 0, 0, 0};
     unsigned char sense_b[SENSE_BUFF_LEN];
@@ -1741,8 +1810,8 @@ int sg_ll_start_stop_unit(int sg_fd, int immed, int power_cond,
     if (verbose) {
         fprintf(sg_warnings_str, "    Start stop unit command:");
         for (k = 0; k < (int)sizeof(ssuBlk); ++k)
-                fprintf (stderr, " %02x", ssuBlk[k]);
-        fprintf(stderr, "\n");
+                fprintf (sg_warnings_str, " %02x", ssuBlk[k]);
+        fprintf(sg_warnings_str, "\n");
     }
     io_hdr.interface_id = 'S';
     io_hdr.cmd_len = sizeof(ssuBlk);
@@ -1766,21 +1835,19 @@ int sg_ll_start_stop_unit(int sg_fd, int immed, int power_cond,
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
         if (verbose)
-            sg_chk_n_print3("Start stop unit", &io_hdr);
+            sg_chk_n_print3("Start stop unit", &io_hdr, (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
-        if (verbose && io_hdr.resid)
-            fprintf(sg_warnings_str, "    Start stop unit: resid=%d\n",
-                    io_hdr.resid);
         return 0;
     case SG_LIB_CAT_INVALID_OP:
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
-            sg_chk_n_print3("START STOP UNIT", &io_hdr);
+            sg_chk_n_print3("START STOP UNIT", &io_hdr, 1);
         return res;
     default:
-        if (verbose)
-            sg_chk_n_print3("START STOP UNIT command error", &io_hdr);
+        if (noisy || verbose)
+            sg_chk_n_print3("START STOP UNIT command error", &io_hdr,
+                            (verbose > 1));
         return -1;
     }
 }
@@ -1790,7 +1857,7 @@ int sg_ll_start_stop_unit(int sg_fd, int immed, int power_cond,
  * Return of 0 -> success,
  * SG_LIB_CAT_INVALID_OP -> command not supported 
  * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, -1 -> other failure */
-int sg_ll_prevent_allow(int sg_fd, int prevent, int verbose)
+int sg_ll_prevent_allow(int sg_fd, int prevent, int noisy, int verbose)
 {
     int k, res;
     unsigned char pCmdBlk[PREVENT_ALLOW_CMDLEN] = 
@@ -1828,10 +1895,14 @@ int sg_ll_prevent_allow(int sg_fd, int prevent, int verbose)
                 "error: %s\n", safe_strerror(errno));
         return -1;
     }
+    if (verbose > 2)
+        fprintf(sg_warnings_str, "      duration=%u ms\n", io_hdr.duration);
     res = sg_err_category3(&io_hdr);
     switch (res) {
     case SG_LIB_CAT_RECOVERED:
-        sg_chk_n_print3("Prevent allow medium removal", &io_hdr);
+        if (verbose)
+            sg_chk_n_print3("Prevent allow medium removal", &io_hdr,
+                            (verbose > 1));
         /* fall through */
     case SG_LIB_CAT_CLEAN:
         return 0;
@@ -1839,13 +1910,162 @@ int sg_ll_prevent_allow(int sg_fd, int prevent, int verbose)
     case SG_LIB_CAT_ILLEGAL_REQ:
         if (verbose > 1)
             sg_chk_n_print3("Prevent allow medium removal command problem",
-                            &io_hdr);
+                            &io_hdr, 1);
         return res;
     default:
-        if (verbose)
+        if (noisy || verbose)
             sg_chk_n_print3("Prevent allow medium removal command problem",
-                            &io_hdr);
+                            &io_hdr, (verbose > 1));
         return -1;
     }
 }
 
+/* Invokes a SCSI REPORT DEVICE IDENTIFIER command. Return of 0 -> success,
+ * SG_LIB_CAT_INVALID_OP -> Read media serial number not supported,
+ * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, -1 -> other failure */
+int sg_ll_report_dev_id(int sg_fd, void * resp, int mx_resp_len,
+                               int noisy, int verbose)
+{
+    int k, res;
+    unsigned char rdiCmdBlk[MAINTENANCE_IN_CMDLEN] =
+                         {MAINTENANCE_IN_CMD, REPORT_DEVICE_IDENTIFIER_SA,
+                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+    unsigned char sense_b[SENSE_BUFF_LEN];
+    struct sg_io_hdr io_hdr;
+
+    rdiCmdBlk[6] = (mx_resp_len >> 24) & 0xff;
+    rdiCmdBlk[7] = (mx_resp_len >> 16) & 0xff;
+    rdiCmdBlk[8] = (mx_resp_len >> 8) & 0xff;
+    rdiCmdBlk[9] = mx_resp_len & 0xff;
+    if (NULL == sg_warnings_str)
+        sg_warnings_str = stderr;
+    if (verbose) {
+        fprintf(sg_warnings_str, "    Report device identifier cdb: ");
+        for (k = 0; k < MAINTENANCE_IN_CMDLEN; ++k)
+            fprintf(sg_warnings_str, "%02x ", rdiCmdBlk[k]);
+        fprintf(sg_warnings_str, "\n");
+    }
+    memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
+    memset(sense_b, 0, sizeof(sense_b));
+    io_hdr.interface_id = 'S';
+    io_hdr.cmd_len = sizeof(rdiCmdBlk);
+    io_hdr.mx_sb_len = sizeof(sense_b);
+    io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
+    io_hdr.dxfer_len = mx_resp_len;
+    io_hdr.dxferp = resp;
+    io_hdr.cmdp = rdiCmdBlk;
+    io_hdr.sbp = sense_b;
+    io_hdr.timeout = DEF_TIMEOUT;
+
+    if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
+        if (noisy || verbose)
+            fprintf(sg_warnings_str, "report_device_idenifier (SG_IO) error:"
+                    " %s\n", safe_strerror(errno));
+        return -1;
+    }
+    if (verbose > 2)
+        fprintf(sg_warnings_str, "      duration=%u ms\n", io_hdr.duration);
+    res = sg_err_category3(&io_hdr);
+    switch (res) {
+    case SG_LIB_CAT_RECOVERED:
+        if (noisy || verbose)
+            sg_chk_n_print3("Report device identifier", &io_hdr,
+                            (verbose > 1));
+        /* fall through */
+    case SG_LIB_CAT_CLEAN:
+        if (verbose && io_hdr.resid)
+            fprintf(sg_warnings_str, "    report_device_identifier: resid=%d ("
+                    "got %d bytes)\n", io_hdr.resid,
+                    io_hdr.dxfer_len - io_hdr.resid);
+        if (verbose > 2) {
+            k = mx_resp_len - io_hdr.resid;
+            if (k > 0) {
+                fprintf(sg_warnings_str, "    report device identifier: "
+                        "response%s\n", (k > 256 ? ", first 256 bytes" : ""));
+                dStrHex(resp, (k > 256 ? 256 : k), -1);
+            }
+        }
+        return 0;
+    case SG_LIB_CAT_INVALID_OP:
+    case SG_LIB_CAT_ILLEGAL_REQ:
+        if (verbose > 1)
+            sg_chk_n_print3("REPORT DEVICE IDENTIFIER", &io_hdr, 1);
+        return res;
+    default:
+        if (noisy || verbose)
+            sg_chk_n_print3("REPORT DEVICE IDENTIFIER command error",
+                            &io_hdr, (verbose > 1));
+        return -1;
+    }
+}
+
+/* Invokes a SCSI SET DEVICE IDENTIFIER command. Return of 0 -> success,
+ * SG_LIB_CAT_INVALID_OP -> Read media serial number not supported,
+ * SG_LIB_CAT_ILLEGAL_REQ -> bad field in cdb, -1 -> other failure */
+int sg_ll_set_dev_id(int sg_fd, void * paramp, int param_len,
+                     int noisy, int verbose)
+{
+    int k, res;
+    unsigned char sdiCmdBlk[MAINTENANCE_OUT_CMDLEN] =
+                         {MAINTENANCE_OUT_CMD, SET_DEVICE_IDENTIFIER_SA,
+                          0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+    unsigned char sense_b[SENSE_BUFF_LEN];
+    struct sg_io_hdr io_hdr;
+
+    sdiCmdBlk[6] = (param_len >> 24) & 0xff;
+    sdiCmdBlk[7] = (param_len >> 16) & 0xff;
+    sdiCmdBlk[8] = (param_len >> 8) & 0xff;
+    sdiCmdBlk[9] = param_len & 0xff;
+    if (NULL == sg_warnings_str)
+        sg_warnings_str = stderr;
+    if (verbose) {
+        fprintf(sg_warnings_str, "    Set device identifier cdb: ");
+        for (k = 0; k < MAINTENANCE_OUT_CMDLEN; ++k)
+            fprintf(sg_warnings_str, "%02x ", sdiCmdBlk[k]);
+        fprintf(sg_warnings_str, "\n");
+        if ((verbose > 1) && paramp && param_len) {
+            fprintf(sg_warnings_str, "    Set device identifier parameter "
+                    "block:\n");
+            dStrHex(paramp, param_len, -1);
+        }
+    }
+    memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
+    memset(sense_b, 0, sizeof(sense_b));
+    io_hdr.interface_id = 'S';
+    io_hdr.cmd_len = sizeof(sdiCmdBlk);
+    io_hdr.mx_sb_len = sizeof(sense_b);
+    io_hdr.dxfer_direction = SG_DXFER_TO_DEV;
+    io_hdr.dxfer_len = param_len;
+    io_hdr.dxferp = paramp;
+    io_hdr.cmdp = sdiCmdBlk;
+    io_hdr.sbp = sense_b;
+    io_hdr.timeout = DEF_TIMEOUT;
+
+    if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
+        if (noisy || verbose)
+            fprintf(sg_warnings_str, "set_device_idenifier (SG_IO) error:"
+                    " %s\n", safe_strerror(errno));
+        return -1;
+    }
+    if (verbose > 2)
+        fprintf(sg_warnings_str, "      duration=%u ms\n", io_hdr.duration);
+    res = sg_err_category3(&io_hdr);
+    switch (res) {
+    case SG_LIB_CAT_RECOVERED:
+        if (noisy || verbose)
+            sg_chk_n_print3("Set device identifier", &io_hdr, (verbose > 1));
+        /* fall through */
+    case SG_LIB_CAT_CLEAN:
+        return 0;
+    case SG_LIB_CAT_INVALID_OP:
+    case SG_LIB_CAT_ILLEGAL_REQ:
+        if (verbose > 1)
+            sg_chk_n_print3("SET DEVICE IDENTIFIER", &io_hdr, 1);
+        return res;
+    default:
+        if (noisy || verbose)
+            sg_chk_n_print3("SET DEVICE IDENTIFIER command error",
+                            &io_hdr, (verbose > 1));
+        return -1;
+    }
+}
