@@ -41,7 +41,7 @@
  *    and a GPL notice.
  *
  *    Much of the data in this file is derived from SCSI draft standards
- *    found at http://www.t10.org with the "SCSI Primary Commands-3" (SPC-3)
+ *    found at http://www.t10.org with the "SCSI Primary Commands-3" (SPC-4)
  *    being the central point of reference.
  *
  *    Other contributions:
@@ -69,7 +69,7 @@
 #include <scsi/sg.h>
 #include <scsi/sg_lib.h>
 
-static char * version_str = "1.10 20050521";    /* spc-3 rev 23+ */
+static char * version_str = "1.15 20051113";    /* spc-4 rev 02 */
 
 FILE * sg_warnings_str = NULL;        /* would like to default to stderr */
 
@@ -496,6 +496,7 @@ static struct error_info additional[] =
     {0x00,0x1a,"Rewind operation in progress"},
     {0x00,0x1b,"Set capacity operation in progress"},
     {0x00,0x1c,"Verify operation in progress"},
+    {0x00,0x1d,"ATA pass through information available"},
     {0x01,0x00,"No index/sector signal"},
     {0x02,0x00,"No seek complete"},
     {0x03,0x00,"Peripheral device write fault"},
@@ -541,6 +542,9 @@ static struct error_info additional[] =
     {0x0B,0x00,"Warning"},
     {0x0B,0x01,"Warning - specified temperature exceeded"},
     {0x0B,0x02,"Warning - enclosure degraded"},
+    {0x0B,0x03,"Warning - background self-test failed"},
+    {0x0B,0x04,"Warning - background pre-scan failed"},
+    {0x0B,0x05,"Warning - background medium scan failed"},
     {0x0C,0x00,"Write error"},
     {0x0C,0x01,"Write error - recovered with auto reallocation"},
     {0x0C,0x02,"Write error - auto reallocation failed"},
@@ -589,6 +593,7 @@ static struct error_info additional[] =
     {0x11,0x11,"Read error - loss of streaming"},
     {0x11,0x12,"Auxiliary memory read error"},
     {0x11,0x13,"Read error - failed retransmission request"},
+    {0x11,0x14,"Read error - LBA marked bad by application client"},
     {0x12,0x00,"Address mark not found for id field"},
     {0x13,0x00,"Address mark not found for data field"},
     {0x14,0x00,"Recorded entity not found"},
@@ -721,6 +726,7 @@ static struct error_info additional[] =
     {0x2D,0x00,"Overwrite error on update in place"},
     {0x2E,0x00,"Insufficient time for operation"},
     {0x2F,0x00,"Commands cleared by another initiator"},
+    {0x2F,0x01,"Commands cleared by power loss notification"},
     {0x30,0x00,"Incompatible medium installed"},
     {0x30,0x01,"Cannot read medium - unknown format"},
     {0x30,0x02,"Cannot read medium - incompatible format"},
@@ -820,6 +826,7 @@ static struct error_info additional[] =
 
     {0x43,0x00,"Message error"},
     {0x44,0x00,"Internal target failure"},
+    {0x44,0x71,"ATA device failed Set Features"}, /* was 0x44,0xf1 in rev1 */
     {0x45,0x00,"Select or reselect failure"},
     {0x46,0x00,"Unsuccessful soft reset"},
     {0x47,0x00,"SCSI parity error"},
@@ -1041,7 +1048,9 @@ static struct error_info additional[] =
 };
 
 static const char *sense_key_desc[] = {
-    "No Sense",                 /* There is no sense information */
+    "No Sense",                 /* Filemark, ILI and/or EOM; progress
+                                   indication (during FORMAT); power
+                                   condition sensing (REQUEST SENSE) */
     "Recovered Error",          /* The last command completed successfully
                                    but used error correction */
     "Not Ready",                /* The addressed target is not ready */
@@ -1147,6 +1156,8 @@ int sg_get_sense_info_fld(const unsigned char * sensep, int sb_len,
     const unsigned char * ucp;
     unsigned long long ull;
 
+    if (info_outp)
+        *info_outp = 0;
     if (sb_len < 7)
         return 0;
     switch (sensep[0] & 0x7f) {
@@ -1168,7 +1179,7 @@ int sg_get_sense_info_fld(const unsigned char * sensep, int sb_len,
             }
             if (info_outp)
                 *info_outp = ull;
-            return !!(ucp[2] & 0x80);
+            return !!(ucp[2] & 0x80);   /* since spc3r23 should be set */
         } else
             return 0;
     default:
@@ -1213,6 +1224,41 @@ int sg_get_sense_progress_fld(const unsigned char * sensep,
     }
 }
 
+static const char * scsi_pdt_strs[] = {
+    /* 0 */ "disk",
+    "tape",
+    "printer",
+    "processor",        /* often SAF-TE (seldom scanner) device */
+    "write once optical disk",
+    /* 5 */ "cd/dvd",
+    "scanner",
+    "optical memory device",
+    "medium changer",
+    "communications",
+    /* 0xa */ "graphics [0xa]",
+    "graphics [0xb]",
+    "storage array controller",
+    "enclosure services device",
+    "simplified direct access device",
+    "optical card reader/writer device",
+    /* 0x10 */ "bridge controller commands",
+    "object based storage",
+    "automation/driver interface",
+    "0x13", "0x14", "0x15", "0x16", "0x17", "0x18",
+    "0x19", "0x1a", "0x1b", "0x1c", "0x1d", 
+    "well known logical unit",
+    "no physical device on this lu",
+};
+
+char * sg_get_pdt_str(int pdt, int buff_len, char * buff)
+{
+    if ((pdt < 0) || (pdt > 31))
+        snprintf(buff, buff_len, "bad pdt");
+    else
+        snprintf(buff, buff_len, "%s", scsi_pdt_strs[pdt]);
+    return buff;
+}
+
 /* Print descriptor format sense descriptors (assumes sense buffer is
    in descriptor format) */
 static void sg_print_sense_descriptors(const unsigned char * sense_buffer,
@@ -1411,7 +1457,7 @@ static void sg_print_sense_descriptors(const unsigned char * sense_buffer,
 
 /* Print sense information */
 void sg_print_sense(const char * leadin, const unsigned char * sense_buffer,
-                    int sb_len)
+                    int sb_len, int raw_sinfo)
 {
     int len, valid, progress;
     unsigned int info;
@@ -1483,7 +1529,8 @@ void sg_print_sense(const char * leadin, const unsigned char * sense_buffer,
                 else if (info > 0)
                     fprintf(sg_warnings_str, "  Valid=0, Info fld=0x%x [%u] ",
                             info, info);
-            }
+            } else
+                info = 0;
             if (sense_buffer[2] & 0xe0) {
                 if (sense_buffer[2] & 0x80)
                    fprintf(sg_warnings_str, " FMK");
@@ -1495,7 +1542,8 @@ void sg_print_sense(const char * leadin, const unsigned char * sense_buffer,
                    fprintf(sg_warnings_str, " ILI");
                             /* incorrect block length requested */
                 fprintf(sg_warnings_str, "\n");
-            }
+            } else if (valid || (info > 0))
+                fprintf(sg_warnings_str, "\n");
             if ((len >= 14) && sense_buffer[14])
                 fprintf(sg_warnings_str, "  Field replaceable unit code: "
                         "%d\n", sense_buffer[14]);
@@ -1550,7 +1598,7 @@ void sg_print_sense(const char * leadin, const unsigned char * sense_buffer,
     } else {    /* non-extended sense data */
 
          /*
-          * Standard says:
+          * A (very old) Standard says:
           *    sense_buffer[0] & 0200 : address valid
           *    sense_buffer[0] & 0177 : vendor-specific error code
           *    sense_buffer[1] & 0340 : vendor-specific
@@ -1575,9 +1623,10 @@ void sg_print_sense(const char * leadin, const unsigned char * sense_buffer,
                 (sense_buffer[0] >> 4) & 0x07, sense_buffer[0] & 0xf);
         len = 4;
     }
-
-    fprintf(sg_warnings_str, " Raw sense data (in hex):\n");
-    dStrHexErr((const char *)sense_buffer, len);
+    if (raw_sinfo) {
+        fprintf(sg_warnings_str, " Raw sense data (in hex):\n");
+        dStrHexErr((const char *)sense_buffer, len);
+    }
 }
 
 static const char * linux_host_bytes[] = {
@@ -1640,9 +1689,10 @@ void sg_print_driver_status(int driver_status)
 
 /* Returns 1 if no errors found and thus nothing printed; otherwise
    prints error/warning (prefix by 'leadin') and returns 0. */
-static int sg_sense_print(const char * leadin, int scsi_status,
-                          int host_status, int driver_status,
-                          const unsigned char * sense_buffer, int sb_len)
+static int sg_linux_sense_print(const char * leadin, int scsi_status,
+                                int host_status, int driver_status,
+                                const unsigned char * sense_buffer,
+                                int sb_len, int raw_sinfo)
 {
     int done_leadin = 0;
     int done_sense = 0;
@@ -1662,7 +1712,7 @@ static int sg_sense_print(const char * leadin, int scsi_status,
         if (sense_buffer && ((scsi_status == SCSI_CHECK_CONDITION) ||
                              (scsi_status == SCSI_COMMAND_TERMINATED))) {
             /* SCSI_COMMAND_TERMINATED is obsolete */
-            sg_print_sense(0, sense_buffer, sb_len);
+            sg_print_sense(0, sense_buffer, sb_len, raw_sinfo);
             done_sense = 1;
         }
     }
@@ -1687,7 +1737,7 @@ static int sg_sense_print(const char * leadin, int scsi_status,
         fprintf(sg_warnings_str, "\n");
         if (sense_buffer && (! done_sense) &&
             (SG_LIB_DRIVER_SENSE == (SG_LIB_DRIVER_MASK & driver_status)))
-            sg_print_sense(0, sense_buffer, sb_len);
+            sg_print_sense(0, sense_buffer, sb_len, raw_sinfo);
     }
     return 0;
 }
@@ -1741,10 +1791,12 @@ int sg_normalize_sense(const struct sg_io_hdr * hp,
 
 /* Returns 1 if no errors found and thus nothing printed; otherwise
    returns 0. */
-int sg_chk_n_print3(const char * leadin, struct sg_io_hdr * hp)
+int sg_chk_n_print3(const char * leadin, struct sg_io_hdr * hp,
+                    int raw_sinfo)
 {
-    return sg_sense_print(leadin, hp->status, hp->host_status,
-                          hp->driver_status, hp->sbp, hp->sb_len_wr);
+    return sg_linux_sense_print(leadin, hp->status, hp->host_status,
+                                hp->driver_status, hp->sbp, hp->sb_len_wr,
+                                raw_sinfo);
 }
 #endif
 
@@ -1752,12 +1804,14 @@ int sg_chk_n_print3(const char * leadin, struct sg_io_hdr * hp)
    returns 0. */
 int sg_chk_n_print(const char * leadin, int masked_status,
                    int host_status, int driver_status,
-                   const unsigned char * sense_buffer, int sb_len)
+                   const unsigned char * sense_buffer, int sb_len,
+                   int raw_sinfo)
 {
     int scsi_status = (masked_status << 1) & 0x7e;
 
-    return sg_sense_print(leadin, scsi_status, host_status, driver_status,
-                          sense_buffer, sb_len);
+    return sg_linux_sense_print(leadin, scsi_status, host_status,
+                                driver_status, sense_buffer, sb_len,
+                                raw_sinfo);
 }
 
 #ifdef SG_IO
@@ -2003,7 +2057,7 @@ char * safe_strerror(int errnum)
 }
 
 
-/* Note the ASCII-hex output goes to stdout. [All other output from functions
+/* Note the ASCII-hex output goes to stdout. [Most other output from functions
    in this file go to sg_warnings_str (default stderr).] 
    'no_ascii' allows for 3 output types:
        > 0     each line has address then up to 16 ASCII-hex bytes
@@ -2046,8 +2100,6 @@ void dStrHex(const char* str, int len, int no_ascii)
     /* no_ascii>=0, start each line with address (offset) */
     k = sprintf(buff + 1, "%.2x", a);
     buff[k + 1] = ' ';
-    if (bpos >= ((bpstart + (9 * 3))))
-        bpos++;
 
     for (i = 0; i < len; i++) {
         c = *p++;
@@ -2060,10 +2112,10 @@ void dStrHex(const char* str, int len, int no_ascii)
             buff[cpos++] = ' ';
         else {
             if ((c < ' ') || (c >= 0x7f))
-                c='.';
+                c = '.';
             buff[cpos++] = c;
         }
-        if (cpos > (cpstart+15)) {
+        if (cpos > (cpstart + 15)) {
             printf("%.76s\n", buff);
             bpos = bpstart;
             cpos = cpstart;
@@ -2110,6 +2162,123 @@ static void dStrHexErr(const char* str, int len)
     return;
 }
 
+/* Returns 1 when executed on big endian machine; else returns 0.
+   Useful for displaying ATA identify words (which need swapping on a
+   big endian machine). */
+int sg_is_big_endian()
+{
+    union u_t {
+        unsigned short s;
+        unsigned char c[sizeof(unsigned short)];
+    } u;
+
+    u.s = 0x0102;
+    return (u.c[0] == 0x01);     /* The lowest address contains
+                                    the most significant byte */
+}
+
+static unsigned short swapb_ushort(unsigned short u)
+{
+    unsigned short r;
+
+    r = (u >> 8) & 0xff;
+    r |= ((u & 0xff) << 8);
+    return r;
+}
+
+/* Note the ASCII-hex output goes to stdout. [Most other output from functions
+   in this file go to sg_warnings_str (default stderr).] 
+   'no_ascii' allows for 3 output types:
+       > 0     each line has address then up to 8 ASCII-hex 16 bit words
+       = 0     in addition, the ASCI bytes pairs are listed to the right
+       = -1    only the ASCII-hex words are listed (i.e. without address)
+       = -2    only the ASCII-hex words, formatted for "hdparm --Istdin"
+       < -2    same as -1
+   If 'swapb' non-zero then bytes in each word swapped. Needs to be set
+   for ATA IDENTIFY DEVICE response on big-endian machines. */
+void dWordHex(const unsigned short* words, int num, int no_ascii,
+              int swapb)
+{
+    const unsigned short * p = words;
+    unsigned short c;
+    char buff[82];
+    unsigned char upp, low;
+    int a = 0;
+    const int bpstart = 3;
+    const int cpstart = 52;
+    int cpos = cpstart;
+    int bpos = bpstart;
+    int i, k;
+
+    if (num <= 0)
+        return;
+    memset(buff, ' ', 80);
+    buff[80] = '\0';
+    if (no_ascii < 0) {
+        for (k = 0; k < num; k++) {
+            c = *p++;
+            if (swapb)
+                c = swapb_ushort(c);
+            bpos += 5;
+            sprintf(&buff[bpos], "%.4x", (unsigned int)c);
+            buff[bpos + 4] = ' ';
+            if ((k > 0) && (0 == ((k + 1) % 8))) {
+                if (-2 == no_ascii)
+                    printf("%.39s\n", buff +8);
+                else
+                    printf("%.47s\n", buff);
+                bpos = bpstart;
+                memset(buff, ' ', 80);
+            }
+        }
+        if (bpos > bpstart) {
+            if (-2 == no_ascii)
+                printf("%.39s\n", buff +8);
+            else
+                printf("%.47s\n", buff);
+        }
+        return;
+    }
+    /* no_ascii>=0, start each line with address (offset) */
+    k = sprintf(buff + 1, "%.2x", a);
+    buff[k + 1] = ' ';
+
+    for (i = 0; i < num; i++) {
+        c = *p++;
+        if (swapb)
+            c = swapb_ushort(c);
+        bpos += 5;
+        sprintf(&buff[bpos], "%.4x", (unsigned int)c);
+        buff[bpos + 4] = ' ';
+        if (no_ascii) {
+            buff[cpos++] = ' ';
+            buff[cpos++] = ' ';
+            buff[cpos++] = ' ';
+        } else {
+            upp = (c >> 8) & 0xff;
+            low = c & 0xff;
+            if ((upp < 0x20) || (upp >= 0x7f))
+                upp = '.';
+            buff[cpos++] = upp;
+            if ((low < 0x20) || (low >= 0x7f))
+                low = '.';
+            buff[cpos++] = low;
+            buff[cpos++] = ' ';
+        }
+        if (cpos > (cpstart + 23)) {
+            printf("%.76s\n", buff);
+            bpos = bpstart;
+            cpos = cpstart;
+            a += 8;
+            memset(buff, ' ', 80);
+            k = sprintf(buff + 1, "%.2x", a);
+            buff[k + 1] = ' ';
+        }
+    }
+    if (cpos > cpstart)
+        printf("%.76s\n", buff);
+}
+
 /* If the number in 'buf' can be decoded or the multiplier is unknown
    then -1 is returned. Accepts a hex prefix (0x or 0X) or a decimal
    multiplier suffix (as per GNU's dd (since 2002: SI and IEC 60027-2)).
